//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Draw and cache the knot's sections.  See KnotSection.h for more info.

#include <cmath>
#include <cstdlib>
#include <cstring>
#include <ctime>

#include <Bitmap.h>
#include <Debug.h>
#include <GraphicsDefs.h>
#include <Point.h>
#include <Rect.h>
#include <Region.h>
#include <View.h>

#include "KnotSectionCache.h"

KnotSectionCache::KnotSectionCache(
	float edgeLength, float outlineWidth, float fillWidth)
:	inherited(BRect(B_ORIGIN, BPoint(edgeLength, edgeLength)),
				 "KnotSectionCache", B_FOLLOW_NONE, B_WILL_DRAW),
	mSize(edgeLength * 0.5),
	mCenter(mSize, mSize),
	mOutlineWidth(edgeLength * outlineWidth),
	mFillWidth(edgeLength * fillWidth),
	mCanvas(BRect(B_ORIGIN, BPoint(edgeLength, edgeLength)),
			  B_MONOCHROME_1_BIT,
			  true)
{
	// Clear the cache.
	memset(mpCache, 0, sizeof(BBitmap *) * 3*4*4*4*4);

	// Attach to the drawing bitmap.
	mCanvas.AddChild(this);
}

KnotSectionCache::~KnotSectionCache(void)
{
	// Detach from the drawing bitmap.
	mCanvas.RemoveChild(this);
	
	// Free cached sections.
	for (int i = 0; i < 3; ++i) {
		for (int j = 0; j < 4; ++j) {
			for (int k = 0; k < 4; ++k) {
				for (int l = 0; l < 4; ++l) {
					for (int m = 0; m < 4; ++m) {
						delete mpCache[i][j][k][l][m];
						mpCache[i][j][k][l][m] = 0;
					}
				}
			}
		}
	}
}

const BBitmap *
KnotSectionCache::GetKnotSection(const KnotSection &section)
{
	ASSERT(section.typ != N);
	
	// Look up the section in the cache.
	BBitmap *&pCached =
		mpCache
			[section.Type()]
			[section.LeftTop()]
			[section.RightTop()]
			[section.LeftBottom()]
			[section.RightBottom()];
	
	if (!pCached) {
		// Draw the section.
		DrawKnotSection(section);
	
		// Cache the resulting bitmap.
		BRect bounds = mCanvas.Bounds();
		pCached = new BBitmap(bounds, B_MONOCHROME_1_BIT);
		pCached->SetBits(mCanvas.Bits(),
								mCanvas.BytesPerRow() * (bounds.IntegerHeight() + 1),
								0,
								B_MONOCHROME_1_BIT);
	}
	
	return pCached;
}

// Private methods for drawing the sections.
void
KnotSectionCache::DrawKnotSection(const KnotSection &section)
{
	if (mCanvas.Lock()) {
		// Erase the bounds rectangle.
		SetDrawingMode(B_OP_ERASE);
		FillRect(Bounds());
		
		// Draw the section.
		switch (section.Type()) {
		
		case D:
			DiagonalSection(section);
			break;
		
		case H:
			HorizontalSection(section);
			break;
		
		case V:
			VerticalSection(section);
			break;
		
		default:
			DEBUGGER("Illegal section type.");
			break;
		}
		
		Sync();
		mCanvas.Unlock();
	}
}

void
KnotSectionCache::DiagonalSection(const KnotSection &section)
{
	// Draw bottom stroke.
//	SetPenSize(mOutlineWidth);
//	SetDrawingMode(B_OP_ERASE);
//	DiagonalStroke(section.RightTop(), section.LeftBottom(), true);
	SetPenSize(mFillWidth);
	SetDrawingMode(B_OP_COPY);
	DiagonalStroke(section.RightTop(), section.LeftBottom(), true);

	// Draw top stroke.
	SetPenSize(mOutlineWidth);
	SetDrawingMode(B_OP_ERASE);
	DiagonalStroke(section.LeftTop(), section.RightBottom(), false);
	SetPenSize(mFillWidth);
	SetDrawingMode(B_OP_COPY);
	DiagonalStroke(section.LeftTop(), section.RightBottom(), false);

	// Draw corners if needed.
	StrokeCorners(section);
}

void
KnotSectionCache::DiagonalStroke(
	direction from, direction to, bool mirror)
{
	switch (4*from + to) {

	case 4*D + D:
		StrokeDDD(mirror ? 4 : 0);
		break;
	
	case 4*D + H:
		StrokeDHD(mirror ? 6 : 2);
		break;
	
	case 4*D + V:
		StrokeDHD(mirror ? 1 : 5);
		break;
	
	case 4*H + D:
		StrokeDHD(mirror ? 4 : 0);
		break;
	
	case 4*H + H:
		StrokeDHH(mirror ? 4 : 0);
		break;
	
	case 4*H + V:
		StrokeDHV(mirror ? 4 : 0);
		break;
	
	case 4*V + D:
		StrokeDHD(mirror ? 3 : 7);
		break;
	
	case 4*V + H:
		StrokeDHV(mirror ? 6 : 2);
		break;
	
	case 4*V + V:
		StrokeDHH(mirror ? 1 : 5);
		break;	
	
	default:
		DEBUGGER("Illegal diagonal stroke.");
		break;
	}
}

void
KnotSectionCache::HorizontalSection(const KnotSection &section)
{
	// Draw bottom stroke.
//	SetPenSize(mOutlineWidth);
//	SetDrawingMode(B_OP_ERASE);
//	HorizontalStroke(section.LeftBottom(), section.RightBottom(), true);
	SetPenSize(mFillWidth);
	SetDrawingMode(B_OP_COPY);
	HorizontalStroke(section.LeftBottom(), section.RightBottom(), true);

	// Draw top stroke.
//	SetPenSize(mOutlineWidth);
//	SetDrawingMode(B_OP_ERASE);
//	HorizontalStroke(section.LeftTop(), section.RightTop(), false);
	SetPenSize(mFillWidth);
	SetDrawingMode(B_OP_COPY);
	HorizontalStroke(section.LeftTop(), section.RightTop(), false);

	// Draw corners if needed.
	StrokeCorners(section);
}

void
KnotSectionCache::HorizontalStroke(
	direction from, direction to, bool mirror)
{
	switch (4*from + to) {

	case 4*D + D:
		StrokeHDD(mirror ? 2 : 0);
		break;
	
	case 4*D + H:
		StrokeHDH(mirror ? 6 : 0);
		break;
	
	case 4*D + V:
		StrokeHDV(mirror ? 6 : 0);
		break;
	
	case 4*H + D:
		StrokeHDH(mirror ? 2 : 4);
		break;
	
	case 4*H + H:
		StrokeHHH(mirror ? 2 : 0);
		break;
	
	case 4*H + V:
		StrokeHHV(mirror ? 6 : 0);
		break;
	
	case 4*V + D:
		StrokeHDV(mirror ? 2 : 4);
		break;
	
	case 4*V + H:
		StrokeHHV(mirror ? 2 : 4);
		break;
	
	case 4*V + V:
		StrokeHVV(mirror ? 2 : 0);
		break;	
	
	case 4*N + N:
		// Stroke omitted (edge section).
		break;
	
	default:
		DEBUGGER("Illegal horizontal stroke.");
		break;
	}
}

void
KnotSectionCache::VerticalSection(const KnotSection &section)
{
	// Draw right stroke.
//	SetPenSize(mOutlineWidth);
//	SetDrawingMode(B_OP_ERASE);
//	VerticalStroke(section.RightTop(), section.RightBottom(), true);
	SetPenSize(mFillWidth);
	SetDrawingMode(B_OP_COPY);
	VerticalStroke(section.RightTop(), section.RightBottom(), true);

	// Draw left stroke.
//	SetPenSize(mOutlineWidth);
//	SetDrawingMode(B_OP_ERASE);
//	VerticalStroke(section.LeftTop(), section.LeftBottom(), false);
	SetPenSize(mFillWidth);
	SetDrawingMode(B_OP_COPY);
	VerticalStroke(section.LeftTop(), section.LeftBottom(), false);

	// Draw corners if needed.
	StrokeCorners(section);
}

void
KnotSectionCache::VerticalStroke(
	direction from, direction to, bool mirror)
{
	switch (4*from + to) {
	
	case 4*D + D:
		StrokeHDD(mirror ? 3 : 1);
		break;
	
	case 4*D + H:
		StrokeHDV(mirror ? 3 : 7);
		break;
	
	case 4*D + V:
		StrokeHDH(mirror ? 3 : 7);
		break;
	
	case 4*H + D:
		StrokeHDV(mirror ? 5 : 1);
		break;
	
	case 4*H + H:
		StrokeHVV(mirror ? 3 : 1);
		break;
	
	case 4*H + V:
		StrokeHHV(mirror ? 5 : 1);
		break;
	
	case 4*V + D:
		StrokeHDH(mirror ? 5 : 1);
		break;
	
	case 4*V + H:
		StrokeHHV(mirror ? 3 : 7);
		break;
	
	case 4*V + V:
		StrokeHHH(mirror ? 3 : 1);
		break;	
	
	case 4*N + N:
		// Stroke omitted (edge section).
		break;
	
	default:
		DEBUGGER("Illegal vertical stroke.");
		break;
	}
}

void
KnotSectionCache::StrokeCorners(const KnotSection &section)
{
	// Draw corners of the section if needed.
	if (section.LeftTop() == D) {
		SetPenSize(mOutlineWidth);
		SetDrawingMode(B_OP_ERASE);
		StrokeLeftTop();
		SetPenSize(mFillWidth);
		SetDrawingMode(B_OP_COPY);
		StrokeLeftTop();
	}
	if (section.RightBottom() == D) {
		SetPenSize(mOutlineWidth);
		SetDrawingMode(B_OP_ERASE);
		StrokeRightBottom();
		SetPenSize(mFillWidth);
		SetDrawingMode(B_OP_COPY);
		StrokeRightBottom();
	}
}

void
KnotSectionCache::StrokeLeftTop(void)
{
	// The top left crossing stroke of a diagonal.
	StrokeLine(mCenter - BPoint(2*mSize, 0), mCenter - BPoint(0, 2*mSize));
}

void
KnotSectionCache::StrokeRightBottom(void)
{
	// The bottom right crossing stroke of a diagonal.
	StrokeLine(mCenter + BPoint(2*mSize, 0), mCenter + BPoint(0, 2*mSize));
}

inline float
orientAngle(float angle, int orientation)
{
	switch (orientation) {
	case 1:
	case 5:
		angle += 90.0;
		break;

	case 2:
	case 6:
		angle += 180.0;
		break;

	case 3:
	case 7:
		angle += 270.0;
		break;

	default:
		break;
	}
	
	if (orientation > 3) {
		angle = 180.0 - angle;
	}

	return angle;
}

inline BPoint
orientPoint(float x, float y, int orientation)
{
	BPoint point(x, y);
	
	switch (orientation) {
	case 1:
	case 5:
		point.x = y;
		point.y = -x;
		break;

	case 2:
	case 6:
		point.x = -x;
		point.y = -y;
		break;

	case 3:
	case 7:
		point.x = -y;
		point.y = x;
		break;

	default:
		break;
	}
	
	if (orientation > 3) {
		point.x = -point.x;
	}
	
	return point;
}

void
KnotSectionCache::StrokeDDD(int orientation)
{
	// The basic diagonal stroke.
	const BPoint offset = orientPoint(mSize, mSize, orientation);
	StrokeLine(mCenter - offset, mCenter + offset);
}

void
KnotSectionCache::StrokeDHD(int orientation)
{
	// Diagonal stroke, horizontal to diagonal.
	const float radius = fabs(mSize) / 2.0 / (M_SQRT2 - 1.0);
	const BPoint arcOffset = orientPoint(-mSize, radius - mSize*0.5, orientation);
	const float angle = orientAngle(67.5, orientation);
	StrokeArc(mCenter + arcOffset, radius, radius, angle - 25.0, 50.0);
	
	const float strokeOffset = mSize * (1.0 / 2.0 / (2.0 - M_SQRT2) - 1.0) - 1.0;
	StrokeLine(mCenter + orientPoint(strokeOffset, strokeOffset, orientation),
				  mCenter + orientPoint(mSize, mSize, orientation));
}

void
KnotSectionCache::StrokeDHH(int orientation)
{
	// Diagonal stroke, horizontal to horizontal.
	const BPoint radius = orientPoint(mSize * 1.34, mSize * 1.5, orientation);
	const float xRadius = fabs(radius.x);
	const float yRadius = fabs(radius.y);
	const BPoint offset = orientPoint(-mSize, mSize, orientation);
	const float angle = orientAngle(67.5, orientation);
	StrokeArc(mCenter + offset, xRadius, yRadius, angle - 25.0, 50.0);
	StrokeArc(mCenter - offset, xRadius, yRadius, angle + 157.0, 50.0);
}

void
KnotSectionCache::StrokeDHV(int orientation)
{
	// Diagonal stroke, horizontal to vertical.
	const float radius = mSize * 1.5;
	StrokeArc(mCenter + orientPoint(-mSize, mSize, orientation),
				 radius, radius, orientAngle(45.0, orientation) - 48.0, 96.0);
}

void
KnotSectionCache::StrokeHDD(int orientation)
{
	// Horizontal stroke, diagonal to diagonal.
	const BPoint radius =
		orientPoint(mSize, mSize * 0.75, orientation);
	const float xRadius = fabs(radius.x);
	const float yRadius = fabs(radius.y);
	StrokeArc(mCenter + orientPoint(0, -1.25 * mSize, orientation),
				 xRadius, yRadius, orientAngle(270.0, orientation) - 96.0, 192.0);
}

void
KnotSectionCache::StrokeHDH(int orientation)
{
	// Horizontal stroke, diagonal to horizontal.
	const BPoint radius =
		orientPoint(mSize, mSize * 0.75, orientation);
	const float xRadius = fabs(radius.x);
	const float yRadius = fabs(radius.y);
	StrokeArc(mCenter + orientPoint(0, -1.25 * mSize, orientation),
				 xRadius, yRadius, orientAngle(225.0, orientation) - 48.0, 96.0);

	StrokeLine(mCenter + orientPoint(0, -0.5 * mSize, orientation),
				  mCenter + orientPoint(mSize, -0.5 * mSize, orientation));
}

void
KnotSectionCache::StrokeHDV(int orientation)
{
	// Horizontal stroke, diagonal to vertical.
	const BPoint radius =
		orientPoint(mSize, mSize * 0.75, orientation);
	const float xRadius = fabs(radius.x);
	const float yRadius = fabs(radius.y);
	StrokeArc(mCenter + orientPoint(0, -1.25 * mSize, orientation),
				 xRadius, yRadius, orientAngle(225.0, orientation) - 48.0, 96.0);

	const BPoint points[3] = {
		mCenter + orientPoint(0, -0.5 * mSize, orientation),
		mCenter + orientPoint(0.5 * mSize, -0.5 * mSize, orientation),
		mCenter + orientPoint(0.5 * mSize, -mSize, orientation)
	};
	StrokePolygon(points, 3, false);
		  
	const float rad = mSize * 0.5;
	StrokeArc(mCenter + orientPoint(0, -mSize, orientation),
				 rad, rad, orientAngle(315.0, orientation) - 48.0, 96.0);
}

void
KnotSectionCache::StrokeHHH(int orientation)
{
	// The basic horizontal stroke.
	StrokeLine(mCenter + orientPoint(-mSize, -0.5 * mSize, orientation),
				  mCenter + orientPoint(mSize, -0.5 * mSize, orientation));
}

void
KnotSectionCache::StrokeHHV(int orientation)
{
	// Horizontal stroke, horizontal to vertical.
	const BPoint points[3] = {
		mCenter + orientPoint(-mSize, -0.5 * mSize, orientation),
		mCenter + orientPoint(0.5 * mSize, -0.5 * mSize, orientation),
		mCenter + orientPoint(0.5 * mSize, -mSize, orientation)
	};
	StrokePolygon(points, 3, false);

	const float radius = mSize * 0.5;
	StrokeArc(mCenter + orientPoint(0, -mSize, orientation),
				 radius, radius, orientAngle(315.0, orientation) - 48.0, 96.0);
}

void
KnotSectionCache::StrokeHVV(int orientation)
{
	// "Horizontal" stroke, vertical to vertical.
	const float radius = mSize * 0.5;
	StrokeArc(mCenter + orientPoint(0, -mSize, orientation),
				 radius, radius, orientAngle(270.0, orientation) - 96.0, 192.0);
}
